
        INCLUDE \masm32\include\masm32rt.inc
        .586
        INCLUDE \masm32\macros\timers.asm

;--------------------------------------------------------------------------

EnumerateCPUs PROTO

;--------------------------------------------------------------------------

;loop counts for timing macros

LOOP_COUNT1 = 42900000
LOOP_COUNT2 = 83000000
LOOP_COUNT3 = 60000000

;--------------------------------------------------------------------------

        .CONST

VTable      dw     8191
            dw     4095
            dw     1023
            dw     255
            dw     -255
            dw     -1023
            dw     -4095
            dw     -8191

Recip       real10 3.0517578125e-5  ;1/32768

        .DATA?

OutValReal4 real4  ?
InVal16s    dw     ?

;--------------------------------------------------------------------------

        .CODE

_main   PROC

;enumerate CPUs
;this function returns the affinity mask of
;the first core that supports CPUID and RDTSC

        INVOKE  EnumerateCPUs
        or      eax,eax
        jnz     Time00

        print   chr$('RDTSC not supported on this processor'),13,10
        jmp     Exit00

;restrict execution to a single core for RDTSC values

Time00: push    eax
        INVOKE  GetCurrentProcess
        pop     edx
        INVOKE  SetProcessAffinityMask,
                eax,
                edx

;---------------------------------------------------------

;Aging n00b/bit-twiddler guy new method

        print   chr$(13,10,'Aging n00b/bit-twiddler guy new method:'),13,10
        mov     edx,offset VTable
        mov     ecx,8

Time01: push    ecx
        movsx   eax,word ptr [edx]
        mov     InVal16s,ax
        push    edx
        print   sstr$(eax),58,9
        mov     ecx,3

Time02: push    ecx

        counter_begin LOOP_COUNT1,HIGH_PRIORITY_CLASS

        movzx   eax,word ptr InVal16s
        or      ax,ax
        jz      Scale2

        mov     cx,16
        jns     Scale0

        neg     ax
        mov     ch,1

Scale0: bsr     dx,ax
        cmp     dl,7
        ja      Scale1

        xchg    al,ah
        mov     cl,8

Scale1: sub     cl,dl
        add     dl,7Fh
        mov     dh,ch
        shl     eax,cl
        shr     edx,1
        rcr     ax,1
        xchg    al,dl
        bswap   eax
        mov     ah,dl

Scale2: mov     OutValReal4,eax

        counter_end
        print   str$(eax),9

        pop     ecx
        dec     ecx
        jz      Time03

        jmp     Time02

Time03: print   chr$('clock cycles'),13,10
        pop     edx
        add     edx,2
        pop     ecx
        dec     ecx
        jz      Time04

        jmp     Time01

;---------------------------------------------------------

;Aging newbie/FPU guy method (resident multiplier)

Time04: print   chr$(13,10,'Aging newbie/FPU guy method ',40,'resident multiplier',41,':'),13,10
        mov     edx,offset VTable
        mov     ecx,8

Time05: push    ecx
        movsx   eax,word ptr [edx]
        mov     InVal16s,ax
        push    edx
        print   sstr$(eax),58,9
        mov     ecx,3

Time06: push    ecx
        finit
        fld real10 ptr Recip

        counter_begin LOOP_COUNT2,HIGH_PRIORITY_CLASS

        fild word ptr InVal16s
        fmul    st,st(1)
        fstp real4 ptr OutValReal4

        counter_end
        print   str$(eax),9

        pop     ecx
        dec     ecx
        jz      Time07

        jmp     Time06

Time07: print   chr$('clock cycles'),13,10
        pop     edx
        add     edx,2
        pop     ecx
        dec     ecx
        jz      Time08

        jmp     Time05

;---------------------------------------------------------

;Aging newbie/FPU guy method (non-resident multiplier)

Time08: print   chr$(13,10,'Aging newbie/FPU guy method ',40,'non-resident multiplier',41,':'),13,10
        mov     edx,offset VTable
        mov     ecx,8

Time09: push    ecx
        movsx   eax,word ptr [edx]
        mov     InVal16s,ax
        push    edx
        print   sstr$(eax),58,9
        mov     ecx,3

Time10: push    ecx
        finit

        counter_begin LOOP_COUNT3,HIGH_PRIORITY_CLASS

        fld real10 ptr Recip
        fild word ptr InVal16s
        fmul    st,st(1)
        fstp real4 ptr OutValReal4
        ffree   st

        counter_end
        print   str$(eax),9

        pop     ecx
        dec     ecx
        jz      Time11

        jmp     Time10

Time11: print   chr$('clock cycles'),13,10
        pop     edx
        add     edx,2
        pop     ecx
        dec     ecx
        jz      Exit00

        jmp     Time09

;---------------------------------------------------------

Exit00: inkey   "Press any key to exit"
        exit

_main   ENDP

;--------------------------------------------------------------------------

        OPTION  PROLOGUE:NONE
        OPTION  EPILOGUE:NONE

EnumerateCPUs PROC

;Enumerate processor cores - by DednDave
;this is a preliminary version - to be updated soon
;I was very careful to leave a lot of room for improvement - lol

;Call With: nothing (future versions will have mode controls)

;  Returns: displays a line of text describing total system cores
;           displays a line of text for each CPU package
;           eax = mask for first core that supports CPUID and RDTSC
;           DF = 0 (up)
;           ebx, ebp, esi, edi preserved

;preserve registers and initialize stack frame

;[ebp] = mask for first core that supports CPUID and RDTSC (init = 0)
;[ebp-4] = CPU number (init = 0)
;[ebp-8] = rotated affinity mask for testing (init = 1)
;[ebp-12] = current process handle
;[ebp-16] = system affinity mask
;[ebp-20] = process affinity mask/cpuid[1]eax
;[ebp-24] = total system cores
;[ebp-28] = cpuid[1]edx
;[ebp-32] = cpuid[1]ecx
;[ebp-36] = cpuid[1]ebx

        push    ebx
        push    esi
        push    edi
        push    ebp
        xor     eax,eax
        push    eax
        mov     ebp,esp
        push    eax
        inc     eax
        push    eax
        sub     esp,28

;get and store current process handle

        INVOKE  GetCurrentProcess
        mov     [ebp-12],eax

;get and store affinity masks

        lea     ecx,[ebp-20]
        lea     edx,[ebp-16]
        INVOKE  GetProcessAffinityMask,
                eax,
                ecx,
                edx

;enumerate total system cores

        print   chr$('Total System Processor Cores: ')
        mov     eax,32
        mov     edx,[ebp-16]
        mov     ecx,eax

Ecpus0: shr     edx,1
        jc      Ecpus1

        dec     eax

Ecpus1: dec     ecx
        jnz     Ecpus0

        mov     [ebp-24],eax
        print   str$(eax),13,10

;enumerate individual CPUs

Ecpus2: print   chr$('CPU ')
        print   str$(dword ptr [ebp-4]),58,32

;restrict execution to the selected core

        INVOKE  SetProcessAffinityMask,
                dword ptr [ebp-12],
                dword ptr [ebp-8]

;verify CPUID supported

        pushfd
        pop     eax
        mov     edx,eax             ;edx=original
        xor     eax,200000h         ;eax=toggled
        push    eax
        popfd                       ;eflags=new
        pushfd
        pop     eax                 ;eax=new
        push    edx
        popfd                       ;eflags=original
        xor     eax,edx             ;eax=toggle result
        test    eax,200000h
        jnz     Ecpus3

        print   chr$('Unknown Cores: 1'),13,10
        mov     ecx,1
        jmp     EcpusR

Ecpus3: mov     eax,1
        cpuid
        mov     [ebp-20],eax
        mov     [ebp-36],ebx
        mov     [ebp-28],edx
        mov     [ebp-32],ecx
        mov     eax,80000000h
        cpuid
        cmp     eax,80000004h
        jae     Ecpus5

;if no brand string, we show fam/mod/xfam/xmod/type/step

Ecpus4: print   chr$('Fam ')
        mov     eax,[ebp-20]
        shr     eax,8
        and     eax,0Fh
        print   str$(eax),32
        print   chr$('Mod ')
        mov     eax,[ebp-20]
        shr     eax,4
        and     eax,0Fh
        print   str$(eax),32
        print   chr$('xFam ')
        mov     eax,[ebp-20]
        shr     eax,20
        and     eax,0FFh
        print   str$(eax),32
        print   chr$('xMod ')
        mov     eax,[ebp-20]
        shr     eax,16
        and     eax,0Fh
        print   str$(eax),32
        print   chr$('Type ')
        mov     eax,[ebp-20]
        shr     eax,12
        and     eax,3
        print   str$(eax),32
        print   chr$('Step ')
        mov     eax,[ebp-20]
        and     eax,0Fh
        print   str$(eax),32
        jmp     Ecpus8

;brand string

Ecpus5: xor     eax,eax
        push    eax
        mov     eax,80000004h
        cpuid
        push    edx
        push    ecx
        push    ebx
        push    eax
        mov     eax,80000003h
        cpuid
        push    edx
        push    ecx
        push    ebx
        push    eax
        mov     eax,80000002h
        cpuid
        push    edx
        push    ecx
        push    ebx
        push    eax

;find the end

        cld
        xor     eax,eax
        mov     edi,esp
        mov     ecx,49
        repnz   scasb
        dec     edi
        dec     edi

;strip spaces off the end

        std
        mov     al,32
        lea     ecx,[edi+1]
        sub     ecx,esp
        repz    scasb
        jz      Ecpus6

        inc     edi

;terminate it with a space and a null

Ecpus6: mov     [edi+1],ax
        lea     esi,[edi+2]
        cld
        mov     edi,esp
        mov     ecx,48
        repz    scasb
        dec     edi

;strip out (r)'s, (tm)'s, and extra spaces
;(code to be added later)



;if less than 2 characters, give them the family/model string instead
;edi = offset of string
;esi = offset of terminator

        sub     esi,edi
        cmp     esi,2
        jbe     Ecpus4

Ecpus7: print   edi
        add     esp,52

;mmx, sse extensions
;(ia32/ia64/EM64T/AMD64 to be added later)
;(possibly HTT and SAHF if line length permits)

Ecpus8: xor     eax,eax
        cpuid
        mov     edx,16800010h
        and     edx,[ebp-28]
        mov     eax,80201h
        cmp     ecx,6C65746Eh   ;is it Intel ?
        jnz     Ecpus9

        or      eax,100000h

Ecpus9: and     eax,[ebp-32]
        or      edx,eax
        cmp     ecx,444D4163h   ;is it AMD ?
        jnz     EcpusA

        push    edx
        mov     eax,80000001h
        cpuid
        pop     eax
        and     ecx,840h
        and     edx,0C0400000h
        or      edx,eax
        or      edx,ecx

EcpusA: mov     ebx,edx
        shr     ebx,18
        mov     ax,dx
        and     ax,0A41h
        and     dl,10h
        or      ebx,eax
        shr     edx,1
        or      bl,dl

;at this point, 14 meaningful bits in BX:

;  Bit Source      Meaning
;   0  0_1[ecx:00] SSE3
;   1  0_1[ecx:19] SSE4.1
;   2  0_1[ecx:20] SSE4.2  (Intel only)
;   3  0_1[edx:04] RDTSC
;   4  8_1[edx:22] MMX+    (AMD only)
;   5  0_1[edx:23] MMX
;   6  8_1[ecx:06] SSE4a   (AMD only)
;   7  0_1[edx:25] SSE
;   8  0_1[edx:26] SSE2
;   9  0_1[ecx:09] SSSE3
;  10  0_1[edx:10] HTT
;  11  8_1[ecx:11] SSE5    (AMD only)
;  12  8_1[edx:30] 3DNow!+ (AMD only)
;  13  8_1[edx:31] 3DNow!  (AMD only)
;  14  0
;  15  0

;mmx extensions

        test    bl,10h
        jz      EcpusB

        print   chr$('MMX+ ')
        jmp short EcpusC

EcpusB: test    bl,20h
        jz      EcpusP          ;if no mmx, no sse or 3dnow either

        print   chr$('MMX ')

;sse extensions

EcpusC: test    bh,8
        jz      EcpusD

        print   chr$('SSE5 ')
        jmp short EcpusK

EcpusD: test    bl,40h
        jz      EcpusE

        print   chr$('SSE4a ')
        jmp short EcpusK

EcpusE: test    bl,4
        jz      EcpusF

        print   chr$('SSE4.2 ')
        jmp short EcpusP        ;if sse4.2 it's Intel, so no 3dnow

EcpusF: test    bl,2
        jz      EcpusG

        print   chr$('SSE4.1 ')
        jmp short EcpusK

EcpusG: test    bh,2
        jz      EcpusH

        print   chr$('SSSE3 ')
        jmp short EcpusK

EcpusH: test    bl,1
        jz      EcpusI

        print   chr$('SSE3 ')
        jmp short EcpusK

EcpusI: test    bh,1
        jz      EcpusJ

        print   chr$('SSE2 ')
        jmp short EcpusK

EcpusJ: test    bl,80h
        jz      EcpusK

        print   chr$('SSE ')

;3dnow extensions

EcpusK: test    bh,10h
        jz      EcpusL

        print   chr$('3DNow!+ ')
        jmp short EcpusP

EcpusL: test    bh,20h
        jz      EcpusP

        print   chr$('3DNow! ')

;count cores in package

EcpusP: mov     ecx,1
        test    bh,4
        jz      EcpusQ

        mov     ecx,[ebp-36]
        shr     ecx,16
        mov     ch,0

EcpusQ: push    ecx
        print   chr$('Cores: ')
        mov     eax,[esp]
        print   str$(eax),13,10
        pop     ecx

;next core

EcpusR: mov     eax,[ebp-8]

EcpusS: shl     eax,1
        jc      EcpusT

        mov     edx,[ebp-16]
        cmp     eax,edx
        ja      EcpusT

        test    eax,edx
        jz      EcpusS

        dec     ecx
        jnz     EcpusS

        mov     [ebp-8],eax
        inc dword ptr [ebp-4]
        jmp     Ecpus2

;find first core that supports rdtsc
;(temporary - this will be incorporated into the code above later on)

EcpusT: mov     eax,1
        mov     [ebp-8],eax

;restrict execution to the selected core

EcpusU: INVOKE  SetProcessAffinityMask,
                dword ptr [ebp-12],
                eax

;verify CPUID supported

        pushfd
        pop     eax
        mov     edx,eax             ;edx=original
        xor     eax,200000h         ;eax=toggled
        push    eax
        popfd                       ;eflags=new
        pushfd
        pop     eax                 ;eax=new
        push    edx
        popfd                       ;eflags=original
        xor     eax,edx             ;eax=toggle result
        test    eax,200000h
        jz      EcpusW

;test for rdtsc

        mov     eax,1
        cpuid
        test    dl,10h
        jz      EcpusW

;test the counter- give it 2 chances to pass

        mov     ecx,2

EcpusV: push    ecx
        xor     eax,eax
        cpuid
        rdtsc
        push    eax
        xor     eax,eax
        cpuid
        rdtsc
        pop     edx
        pop     ecx
        cmp     eax,edx
        ja      EcpusY

        dec     ecx
        jnz     EcpusV

;next core

EcpusW: mov     eax,[ebp-8]

EcpusX: shl     eax,1
        jc      EcpusZ

        mov     edx,[ebp-16]
        cmp     eax,edx
        ja      EcpusZ

        test    eax,edx
        jz      EcpusX

        mov     [ebp-8],eax
        jmp     EcpusU

;mark it as usable and exit

EcpusY: mov     eax,[ebp-8]
        mov     [ebp],eax

;exit the routine

EcpusZ: INVOKE  SetProcessAffinityMask,
                dword ptr [ebp-12],
                dword ptr [ebp-16]
        mov     esp,ebp
        pop     eax
        pop     ebp
        pop     edi
        pop     esi
        pop     ebx
        ret

EnumerateCPUs ENDP

        OPTION  PROLOGUE:PrologueDef
        OPTION  EPILOGUE:EpilogueDef

;--------------------------------------------------------------------------

        END     _main
